/* Directory Contents Utility
 *
 * Version 1.2
 * 1/14/2021
 *
 * dr.c
 */

#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/md5.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>

#include "color.h"

#define PROG_HELP \
"Directory Contents Utility: List directory contents\n\
\n\
dr [-H] [-htasp] [dir ...] [file ...]\n\
\n\
\t[-H] - Display program syntax and options\n\
\t[-h] - Do not display MD5 hashes\n\
\t[-t] - Do not display total size\n\
\t[-a] - Also display hidden files\n\
\t[-s] - Do not display individual file sizes\n\
\t[-p] - Do not display file permissions\n"

#define OPT_HELP	'H'
#define OPT_NO_MD5	'h'
#define OPT_NO_TOTAL	't'
#define OPT_ALL_FILES	'a'
#define OPT_NO_SIZES	's'
#define OPT_NO_PERMS    'p'

int	g_iTotalSize;
double	g_TotalSize;
char g_lin[] = "";

int files = 0;
int bins  = 0;
int dirs  = 0;

char options[] = "";

enum
{
	_NONE,
	_FILE,
	_BIN,
	_DIR
};

struct mode_bit
{
	unsigned int regular	: 1;
	unsigned int dir	: 1;
	unsigned int chr	: 1;
	unsigned int setuid	: 1;
	unsigned int usr_read	: 1;
	unsigned int usr_write	: 1;
	unsigned int usr_exec	: 1;
	unsigned int grp_read	: 1;
	unsigned int grp_write 	: 1;
	unsigned int grp_exec 	: 1;
	unsigned int oth_read 	: 1;
	unsigned int oth_write 	: 1;
	unsigned int oth_exec 	: 1;
};

int inopts(char opt)
{
	static int sr;
	sr = 0;
	int z;

	for (z = 0; z <= strlen(options); z++)
	{
		if (options[z] == opt)
		{
			sr = 1;
			break;
		} else {
			continue;
		}
	}

	return sr;
}

int ftype(char file[])
{
	struct stat fstat;
	int s = stat(file, &fstat);
	if (s != 0)
	{
		return _NONE;
	}
	if (S_ISREG (fstat.st_mode)) 
	{
		if (((fstat.st_mode & S_IXUSR) || (fstat.st_mode & S_IXGRP) || (fstat.st_mode & S_IXOTH)) && (!S_ISDIR (fstat.st_mode)))
		{
			return _BIN;
		}
		return _FILE;
	} else if (S_ISDIR (fstat.st_mode)) 
	{
		return _DIR;
	} else {
		return _NONE;
	}
}

char * filesize(char tfile[])
{
	if (inopts(OPT_NO_SIZES))
	{
		return "";
	}
	if (ftype(tfile) == _NONE)
	{
		return "(file not found)";
	}
	char sizestr[24];
	static char _sizestr[26];
	double fsize;
	double kb;
	double mb;
	struct stat file;
	stat(tfile, &file);
	int _fsize = file.st_size;
	g_iTotalSize += _fsize;
	fsize = (double)_fsize;
	kb = fsize / 1024.0;
	mb = fsize / 1048576.0;
	if (fsize < 1024.0)
	{
		sprintf(sizestr, "%i bytes", _fsize);
	} else if (fsize >= 1024.0 && fsize < 1048576.0) {
		sprintf(sizestr, "%.2f KB", kb);
	} else if (fsize >= 1048576.0) {
		sprintf(sizestr, "%.2f MB", mb);
	}
	sprintf(_sizestr, "(%s)", sizestr);
	return _sizestr;
}

char * total_size(void)
{
	char tsizestr[24];
	static char rsizestr[28];
	double fsize;
	double kb;
	double mb;
	fsize = (double)g_iTotalSize;
	kb = fsize / 1024.0;
	mb = fsize / 1048576.0;
	if (fsize < 1024.0)
	{
		sprintf(tsizestr, "%i bytes", g_iTotalSize);
	} else if (fsize >= 1024.0 && fsize < 1048576.0) {
		sprintf(tsizestr, "%.2f KB", kb);
	} else if (fsize >= 1048576.0) {
		sprintf(tsizestr, "%.2f MB", mb);
	}
	sprintf(rsizestr, "(%s)", tsizestr);
	return rsizestr;
}

unsigned long fdsize(int fd)
{
	struct stat statbuf;
	if (fstat(fd, &statbuf) < 0)
	{
		return -1;
	}
	return statbuf.st_size;
}

char * md5file(char fname[])
{
	if (ftype(fname) == _DIR)
	{
		return "(directory)                     ";
	}

	int i;
	int file_descript;
	unsigned long file_size;
	char* file_buffer;

	unsigned char result[MD5_DIGEST_LENGTH];
	static char md5f_retv[33];

	file_descript = open(fname, O_RDONLY);
	if (file_descript < 0)
	{
		return "(file access error)             ";
	}

	file_size = fdsize(file_descript);

	file_buffer = mmap(0, file_size, PROT_READ, MAP_SHARED, file_descript, 0);
	MD5((unsigned char*) file_buffer, file_size, result);
	munmap(file_buffer, file_size); 

	for(i = 0; i < MD5_DIGEST_LENGTH; i++)
	{
		sprintf(&md5f_retv[i*2], "%02x", result[i]);
	}

	return md5f_retv;
}

struct mode_bit mode;

void init_mode_struct(char file[])
{
	struct stat st;
	int t = stat(file, &st);
	if (t != 0)
	{
		return;
	}

	if (S_IFREG & st.st_mode)
		mode.regular = 1;

	if (S_IFDIR & st.st_mode)
		mode.dir = 1;

	if (S_IFCHR & st.st_mode)
		mode.chr = 1;

	if (S_ISUID & st.st_mode)
		mode.setuid = 1;

	if (S_IRUSR & st.st_mode)
		mode.usr_read = 1;

	if (S_IWUSR & st.st_mode)
		mode.usr_write = 1;

	if (S_IXUSR & st.st_mode)
		mode.usr_exec = 1;

	if (S_IRGRP & st.st_mode)
		mode.grp_read = 1;

	if (S_IWGRP & st.st_mode)
		mode.grp_write = 1;

	if (S_IXGRP & st.st_mode)
		mode.grp_exec = 1;

	if (S_IROTH & st.st_mode)
		mode.oth_read = 1;

	if (S_IWOTH & st.st_mode)
		mode.oth_write = 1;

	if (S_IXOTH & st.st_mode)
		mode.oth_exec = 1;

	return;
}

void clear_mode_struct(void)
{
	mode.regular	= 0;
	mode.dir	= 0;
	mode.chr	= 0;
	mode.setuid	= 0;
	mode.usr_read	= 0;
	mode.usr_write	= 0;
	mode.usr_exec	= 0;
	mode.grp_read	= 0;
	mode.grp_write	= 0;
	mode.grp_exec	= 0;
	mode.oth_read	= 0;
	mode.oth_write	= 0;
	mode.oth_exec	= 0;
	return;
}

char * modestr(char file[])
{
	if (inopts(OPT_NO_PERMS))
	{
	return "";
	}
	int i = 1;
	static char rstr[12];

	init_mode_struct(file);

	rstr[0] = ' ';

	if (mode.dir && !mode.regular)
	{
		rstr[i] = 'd';
	} else if (mode.chr && !mode.regular) {
		rstr[i] = 'c';
	} else if (mode.regular) {
		rstr[i] = '-';
	}
	i++;
	
	if (mode.usr_read)
	{
		rstr[i] = 'r';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.usr_write)
	{
		rstr[i] = 'w';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.usr_exec && !mode.setuid)
	{
		rstr[i] = 'x';
	} else if (mode.usr_exec && mode.setuid) {
		rstr[i] = 's';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.grp_read)
	{
		rstr[i] = 'r';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.grp_write)
	{
		rstr[i] = 'w';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.grp_exec && !mode.setuid)
	{
		rstr[i] = 'x';
	} else if (mode.grp_exec && mode.setuid) {
		rstr[i] = 's';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.oth_read)
	{
		rstr[i] = 'r';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.oth_write)
	{
		rstr[i] = 'w';
	} else {
		rstr[i] = '-';
	}
	i++;

	if (mode.oth_exec)
	{
		rstr[i] = 'x';
	} else {
		rstr[i] = '-';
	}
	i++;
    
    rstr[i] = ' ';
	
	clear_mode_struct();

	return rstr;
}

int dir_contents(char directory[]) 
{
	struct dirent *de;
	char fdir[128];

	sprintf(fdir, "%s/%s", getenv("PWD"), directory);

	DIR *dptr;
	dptr = opendir(fdir);

	if (!dptr)
	{
		dptr = opendir(directory);
	}

	if (!dptr)
	{
		printf("Failed to load contents of %s: %s\n", directory, strerror(errno));
		return 1;
	}

	while ((de = readdir(dptr)) != NULL)
	{
		if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) 
		{
			if (de->d_name[0] == '.' && !inopts(OPT_ALL_FILES))
			{
				continue;
			}

			if (ftype(de->d_name) != _NONE)
			{
				if (!inopts(OPT_NO_MD5))
				{
					printf("\t%s%sMD5: %s %s%s%s%s %s\n", BOLD, WHITE, md5file(de->d_name), RESET, modestr(de->d_name), de->d_name, RESET, filesize(de->d_name));
				} else {
					printf("\t%s%s%s\n", modestr(de->d_name), de->d_name, filesize(de->d_name));
				}
				if (ftype(de->d_name) == _BIN)
				{
					bins++;
					files++;
				} else if (ftype(de->d_name) == _FILE) {
					files++;
				} else if (ftype(de->d_name) == _DIR) {
					dirs++;
				}
			}
		}
	}
	
	closedir(dptr);

	if (!inopts(OPT_NO_TOTAL))
	{
		printf("%d files (%d executable) %s\n%d directories\n", files, bins, total_size(), dirs);
	}

	return 0; 
}

int main(int argc, char *argv[])
{
	int _files = 0;
	int _bins = 0;
	int has_dir = 0;
	int p = 1;
	int r;
	int i;

	if (argc < 2 || (argc == 2 && argv[1][0] == '-'))
	{
		if (argc > 1)
		{
			for (r = 1; r < strlen(argv[1]); r++)
			{
				options[r - 1] = argv[1][r];
			}
		}

		if (inopts(OPT_HELP))
		{
			printf(PROG_HELP);
			exit(0);
		}

		printf("%s:\n", getenv("PWD"));
		dir_contents(("."));
		exit(0);
	}

	if (argc >= 2 && argv[1][0] == '-')
	{
		p = 2;

		for (r = 1; r < strlen(argv[1]); r++)
		{
			options[r - 1] = argv[1][r];
		}
	}

	for (i = p; i <= (argc - 1); i++)
	{
		if (ftype(argv[i]) == _NONE)
		{
			printf("%s: %s: No such file or directory\n", argv[0], argv[i]);
			continue;
		}

		if (!inopts(OPT_ALL_FILES) && argv[i][0] == '.')
		{
			continue;
		}

		if (ftype(argv[i]) == _DIR)
		{
			printf("%s:\n", argv[i]);
			chdir(argv[i]);
			dir_contents(argv[i]);
			has_dir++;
		} else if (ftype(argv[i]) != _NONE) {
			if (!inopts(OPT_NO_MD5))
			{
				printf("\t%s%sMD5: %s %s%s%s%s %s\n", BOLD, WHITE, md5file(argv[i]), RESET, modestr(argv[i]), argv[i], RESET, filesize(argv[i]));
			} else {
				printf("\t%s%s %s%s\n", modestr(argv[i]), argv[i], RESET, filesize(argv[i]));
			}

			if (ftype(argv[i]) == _BIN)
			{
				_bins++;
				_files++;
			} else if (ftype(argv[i]) == _FILE) {
				_files++;
			}
		}
	}
	
	if (!inopts(OPT_NO_TOTAL))
	{
		if (!has_dir)
		{
			printf("%d files (%d executable) %s\n", _files, _bins, total_size());
		}
	}
	exit(0);
}
