/* Message Server Utility
 *
 * Version 1.0
 * 1/1/2021
 *
 * msgsrv.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define PROG_HELP \
"Message Server Utility: Simple telnet server which outputs text upon connection\n\
\n\
msgsrv [-H] [-f file] [-m \"mesg\"] <-p port> &\n\
\n\
\t[-H] - Display program syntax and options\n\
\t[-f] - Output text of [file] on connection\n\
\t[-m] - Output [message] on connection\n\
\t<-p> - Set TCP port number to listen on for connections as <port>\n\
\n\
\t[mesg] should be enclosed in single quotes or double quotes.\n\
\tOptions -f and -m cannot be used together.\n"

#define OPT_HELP	'H'
#define OPT_PORT	'p'
#define OPT_FILE	'f'
#define OPT_MESG	'm'

/* #define MSG_CONNECT	"* Connected *\n" */
/* #define MSG_DISCONN	"* Connection closed *\n" */

#define MESG_LEN	2048

char g_mesg[MESG_LEN];
char g_opts[8];
char g_file[64];
int g_port = 0;

enum
{
	_NONE,
	_FILE,
	_BIN,
	_DIR
};

int inopts(char opt)
{
	static int sr;
	sr = 0;
	int z;

	for (z = 0; z <= strlen(g_opts); z++)
	{
		if (g_opts[z] == opt)
		{
			sr = 1;
			break;
		} else {
			continue;
		}
	}

	return sr;
}

int ftype(char file[])
{
	struct stat fstat;
	int s = stat(file, &fstat);
	if (s != 0)
	{
		return _NONE;
	}
	if (S_ISREG (fstat.st_mode)) 
	{
		if (((fstat.st_mode & S_IXUSR) || (fstat.st_mode & S_IXGRP) || (fstat.st_mode & S_IXOTH)) && (!S_ISDIR (fstat.st_mode)))
		{
			return _BIN;
		}
		return _FILE;
	} else if (S_ISDIR (fstat.st_mode)) 
	{
		return _DIR;
	} else {
		return _NONE;
	}
}

int vsend(int conn, char str[])
{
	send(conn, str, strlen(str), 1);
	return 0;
}

int fsend(int conn, char file[])
{
	int x;
	FILE *fptr;
	fptr = fopen(g_file, "r");
	for (x = 0; ; x++)
	{
		if (feof(fptr))
		{
			break;
		}
		g_mesg[x] = fgetc(fptr);
	}
	fclose(fptr);
	vsend(conn, g_mesg);
	return 0;
}

int main(int argc, char *argv[])
{
	if (argc > 1 && argv[1][0] == '-' && argv[1][1] == 'H')
	{
		printf(PROG_HELP);
		exit(0);
	}

	if (argc < 5)
	{
		printf("Usage: %s [-H] [-f file] [-m \"mesg\"] <-p port> &\n", argv[0]);
		exit(1);
	}
	
	int x;

	for (x = 1; x <= (argc - 1); x++)
	{
		if (argv[x][0] == '-' && argv[x][1] == 'f')
		{
			strcpy(g_file, argv[x + 1]);
			strcat(g_opts, "f");
			break;
		}
	}

	for (x = 1; x <= (argc - 1); x++)
	{
		if (argv[x][0] == '-' && argv[x][1] == 'p')
		{
			g_port = atoi(argv[x + 1]);
			strcat(g_opts, "p");
			break;
		}
	}

	for (x = 1; x <= (argc - 1); x++)
	{
		if (argv[x][0] == '-' && argv[x][1] == 'm')
		{
			strcpy(g_mesg, argv[x + 1]);
			strcat(g_opts, "m");
			break;
		}
	}

	if (inopts(OPT_FILE) && inopts(OPT_MESG))
	{
		printf("%s: -f and -m cannot be used together\n", argv[0]);
		exit(1);
	}

	if (!inopts(OPT_FILE) && !inopts(OPT_MESG))
	{
		printf("%s: -f or -m cannot be used together\n", argv[0]);
		exit(1);
	}

	if (inopts(OPT_PORT) && g_port <= 0)
	{
		printf("%s: port number not specified\n", argv[0]);
		exit(1);
	} else if (inopts(OPT_FILE) && !strlen(g_file)) {
		printf("%s: -f used without providing file name\n", argv[0]);
		exit(1);
	} else if (inopts(OPT_MESG) && !strlen(g_mesg)) {
		printf("%s: -m used without providing message\n", argv[0]);
		exit(1);
	}

	if (inopts(OPT_FILE))
	{
		if (ftype(g_file) == _NONE)
		{
			printf("%s: %s: No such file or directory\n", argv[0], g_file);
			exit(1);
		} else if (ftype(g_file) == _DIR) {
			printf("%s: %s: Is a directory\n", argv[0], g_file);
			exit(1);
		}
	}

	int opt = 1;
	struct sockaddr_in addr;
	int sock;
	int addrlen = sizeof(addr);
	int sfd = socket(AF_INET, SOCK_STREAM, 0);

	if (sfd == 0)
	{
		perror("socket fd");
		exit(1);
	}

	if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)))
	{
		perror("setsockopt");
		exit(1);
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(g_port);

	if (bind(sfd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
	{
		perror("bind failed");
		exit(1);
	}

	if (listen(sfd, 8))
	{
		perror("listen failed");
		exit(1);
	}

	while (1)
	{
		if ((sock = accept(sfd, (struct sockaddr *)&addr, (socklen_t *)&addrlen)) < 0)
		{
			perror("accept");
			exit(1);
		}
		
		#ifdef MSG_CONNECT
		vsend(sock, MSG_CONNECT);
		#endif /* MSG_CONNECT */

		if (!inopts(OPT_FILE))
		{
			send(sock, g_mesg, strlen(g_mesg), 0);
			send(sock, "\n", 2, 0);
		} else {
			fsend(sock, g_file);
		}

		#ifdef MSG_DISCONN
		vsend(sock, MSG_DISCONN);
		#endif /* MSG_DISCONN */

		close(sock);
	}

	return 0;
}
